package org.javacodegeeks.examples;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import java.util.stream.IntStream;

public class MesozoicExampleMain {

    public static void main(String[] args) throws InterruptedException, ExecutionException {
        // Generate names
        Random r = new Random();
        var names = List.of("Joe", "Monica", "Chandler", "Phoebe", "Rachel", "Ross", "Janice");
        var namesList = IntStream.range(0, 1000000)
            .mapToObj(__ -> names.get(r.nextInt(names.size())))
            .toList();

        int batchSize = 1000;
        int parallelism = (int) Math.ceil(namesList.size() / (double) batchSize);
        System.out.printf("Parallelism is %s \n", parallelism);
        ExecutorService executorService =
            Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors(),
                new ThreadFactory() {
                    @Override
                    public Thread newThread(Runnable runnable) {
                        Thread t = new Thread(runnable);
                        t.setDaemon(true);
                        return t;
                    }
                });

        // Split names into batches
        Map<String, Long> finalCounts = new ConcurrentHashMap<>();
        List<Callable<Void>> tasks = new ArrayList<>();
        for (int i = 0; i < namesList.size(); i += batchSize) {
            int batchStart = i;
            int batchEnd = Math.min((batchStart + batchSize), namesList.size());
            final List<String> batch = namesList.subList(batchStart, batchEnd);
            tasks.add(new CountTask(batch, finalCounts));
        }

        // Wait until tasks are done
        executorService.invokeAll(tasks);


        // Find the max count
        String mostFrequentName = finalCounts.entrySet()
            .stream()
            .max(Map.Entry.comparingByValue())
            .get()
            .getKey();

        System.out.printf("The most frequent name is %s%n", mostFrequentName);
    }

    private static class CountTask implements Callable<Void> {
        private final List<String> batch;
        private final Map<String, Long> finalCounts;

        private CountTask(List<String> batch, Map<String, Long> finalCounts) {
            this.batch = batch;
            this.finalCounts = finalCounts;
        }

        @Override
        public Void call() {
            Map<String, Long> localCounts = new HashMap<>();
            System.out.printf("[%s] Processing batch... \n", Thread.currentThread().getName());
            for (String name : batch) {
                localCounts.compute(name, (n, c) -> c == null ? 1L : c + 1);
            }
            for (Map.Entry<String, Long> stringLongEntry : localCounts.entrySet()) {
                synchronized (finalCounts) {
                    final Long existingCount = finalCounts.get(stringLongEntry.getKey());
                    final var newCount = stringLongEntry.getValue();
                    if (existingCount == null) {
                        finalCounts.put(stringLongEntry.getKey(), newCount);
                    } else {
                        finalCounts.put(stringLongEntry.getKey(), existingCount + newCount);
                    }
                }
            }
            return null;
        }
    }
}
